"""Tests for the `UsbTransport` class."""

import unittest

from adb_shell.exceptions import TcpTimeoutException

try:
    from adb_shell.transport.usb_transport import UsbTransport
except (ImportError, OSError):
    UsbTransport = None

from . import patchers


# pylint: disable=missing-class-docstring, missing-function-docstring
@unittest.skipIf(UsbTransport is None, "UsbTransport could not be imported")
class TestUsbTransport(unittest.TestCase):
    def setUp(self):
        """Create a ``UsbTransport`` and do something...

        """
        self.transport = UsbTransport('TODO', 'TODO')

        if True:
            return

        with patchers.PATCH_CREATE_CONNECTION:
            self.transport.connect()

    def tearDown(self):
        """Close the USB connection."""
        self.transport.close()

    def test_connect_with_timeout(self):
        """TODO

        """
        if True:
            return

        self.transport.close()
        with patchers.PATCH_CREATE_CONNECTION:
            self.transport.connect(transport_timeout_s=1)
            self.assertTrue(True)

    def test_bulk_read(self):
        """TODO

        """
        if True:
            return

        # Provide the `recv` return values
        self.transport._connection._recv = b'TEST1TEST2'

        with patchers.PATCH_SELECT_SUCCESS:
            self.assertEqual(self.transport.bulk_read(5), b'TEST1')
            self.assertEqual(self.transport.bulk_read(5), b'TEST2')

        with patchers.PATCH_SELECT_FAIL:
            with self.assertRaises(TcpTimeoutException):
                self.transport.bulk_read(4)

    def test_bulk_write(self):
        """TODO

        """
        if True:
            return

        with patchers.PATCH_SELECT_SUCCESS:
            self.transport.bulk_write(b'TEST')

        with patchers.PATCH_SELECT_FAIL:
            with self.assertRaises(TcpTimeoutException):
                self.transport.bulk_write(b'FAIL')



import os
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from adb_shell.auth.keygen import keygen
from adb_shell.auth.sign_pythonrsa import PythonRSASigner

from .keygen_stub import open_priv_pub


class TestPythonRSASigner(unittest.TestCase):
    def setUp(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            self.signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')

    def test_sign(self):
        """Test that the ``Sign`` method does not raise an exception."""
        self.signer.Sign(b'notadb')
        self.assertTrue(True)

    def test_get_public_key(self):
        """Test that the ``GetPublicKey`` method works correctly."""
        with patch('{}.open'.format(__name__), open_priv_pub):
            with open('tests/adbkey.pub') as f:
                pub = f.read()

            self.assertEqual(pub, self.signer.GetPublicKey())


class TestPythonRSASignerExceptions(unittest.TestCase):
    def test_value_error(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            with patch('adb_shell.auth.sign_pythonrsa.decoder.decode', return_value=([None, [None]], None)):
                with self.assertRaises(ValueError):
                    keygen('tests/adbkey')
                    self.signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')

    def test_index_error(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            with patch('adb_shell.auth.sign_pythonrsa.decoder.decode', side_effect=IndexError):
                with self.assertRaises(ValueError):
                    keygen('tests/adbkey')
                    self.signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')
                



import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

try:
    from adb_shell.transport.usb_transport import UsbTransport
except (ImportError, OSError):
    UsbTransport = None


class TestUsbImportError(unittest.TestCase):
    def test_import_error(self):
        """Test that the package still works when ``libusb1`` is not installed."""
        from adb_shell import adb_device
        from adb_shell.exceptions import InvalidTransportError

        # TODO: I can't manage to trigger an `ImportError` in adb_device.py
        # self.assertIsNone(adb_device.UsbTransport)

        # In lieu of a real `ImportError`, I'll just set this to None
        with patch("adb_shell.adb_device.UsbTransport", None):
            with self.assertRaises(InvalidTransportError):
                adb_device.AdbDeviceUsb('serial')

    def test_import_successful(self):
        from adb_shell import adb_device

        if UsbTransport is not None:
            # Make sure `UsbTransport` was imported
            with patch("adb_shell.adb_device.UsbTransport", UsbTransport):
                with patch("adb_shell.adb_device.UsbTransport.find_adb", return_value=UsbTransport("TODO", "TODO")):
                    adb_device.AdbDeviceUsb("serial")



import unittest

from adb_shell import constants
from adb_shell.hidden_helpers import _AdbPacketStore, _AdbTransactionInfo


class TestAdbPacketStore(unittest.TestCase):
    def setUp(self):
        self.packet_store = _AdbPacketStore()

    def test_init(self):
        self.assertEqual(len(self.packet_store), 0)
        self.assertFalse((None, None) in self.packet_store)

    def test_contains(self):
        self.assertFalse((None, None) in self.packet_store)

        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")
        self.assertTrue((0, 1) in self.packet_store)
        self.assertTrue((None, 1) in self.packet_store)
        self.assertFalse((None, 0) in self.packet_store)
        self.assertTrue((0, None) in self.packet_store)
        self.assertFalse((1, None) in self.packet_store)
        self.assertFalse((1, 1) in self.packet_store)
        self.assertTrue((None, None) in self.packet_store)

    def test_put(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")
        self.assertTrue((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 1)

        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd2", data=b"data2")
        self.assertTrue((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 1)

        self.packet_store.put(arg0=1, arg1=0, cmd=b"cmd3", data=b"data3")
        self.assertTrue((1, 0) in self.packet_store)
        self.assertEqual(len(self.packet_store), 2)

        self.packet_store.put(arg0=1, arg1=1, cmd=b"cmd4", data=b"data4")
        self.assertTrue((1, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 3)

        self.packet_store.put(arg0=5, arg1=5, cmd=constants.CLSE, data=b"data5")
        self.assertEqual(len(self.packet_store), 3)
        self.assertFalse((5, 5) in self.packet_store)
        
        self.packet_store.put(arg0=5, arg1=1, cmd=constants.CLSE, data=b"data5")
        self.assertEqual(len(self.packet_store), 3)
        self.assertFalse((5, 1) in self.packet_store)

    def test_get(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd2", data=b"data2")
        self.packet_store.put(arg0=1, arg1=0, cmd=b"cmd3", data=b"data3")
        self.packet_store.put(arg0=1, arg1=1, cmd=b"cmd4", data=b"data4")
        self.packet_store.put(arg0=2, arg1=3, cmd=b"cmd5", data=b"data5")
        self.packet_store.put(arg0=4, arg1=5, cmd=b"cmd6", data=b"data6")

        self.assertTrue((0, 1) in self.packet_store)
        cmd1, arg0, arg1, data1 = self.packet_store.get(arg0=0, arg1=1)
        self.assertEqual(arg0, 0)
        self.assertEqual(arg1, 1)
        self.assertEqual(cmd1, b"cmd1")
        self.assertEqual(data1, b"data1")
        self.assertTrue((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 5)

        self.assertTrue((0, 1) in self.packet_store)
        cmd2, arg0, arg1, data2 = self.packet_store.get(arg0=0, arg1=1)
        self.assertEqual(arg0, 0)
        self.assertEqual(arg1, 1)
        self.assertEqual(cmd2, b"cmd2")
        self.assertEqual(data2, b"data2")
        self.assertFalse((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 4)

        self.assertTrue((1, 0) in self.packet_store)
        cmd3, arg0, arg1, data3 = self.packet_store.get(arg0=1, arg1=0)
        self.assertEqual(arg0, 1)
        self.assertEqual(arg1, 0)
        self.assertEqual(cmd3, b"cmd3")
        self.assertEqual(data3, b"data3")
        self.assertEqual(len(self.packet_store), 3)

        self.assertTrue((1, None) in self.packet_store)
        cmd4, arg0, arg1, data4 = self.packet_store.get(arg0=1, arg1=None)
        self.assertEqual(arg0, 1)
        self.assertEqual(arg1, 1)
        self.assertEqual(cmd4, b"cmd4")
        self.assertEqual(data4, b"data4")
        self.assertEqual(len(self.packet_store), 2)

        self.assertTrue((None, 3) in self.packet_store)
        cmd5, arg0, arg1, data5 = self.packet_store.get(arg0=None, arg1=3)
        self.assertEqual(arg0, 2)
        self.assertEqual(arg1, 3)
        self.assertEqual(cmd5, b"cmd5")
        self.assertEqual(data5, b"data5")
        self.assertEqual(len(self.packet_store), 1)

        self.assertTrue((None, None) in self.packet_store)
        cmd6, arg0, arg1, data6 = self.packet_store.get(arg0=None, arg1=None)
        self.assertEqual(arg0, 4)
        self.assertEqual(arg1, 5)
        self.assertEqual(cmd6, b"cmd6")
        self.assertEqual(data6, b"data6")
        self.assertEqual(len(self.packet_store), 0)

        self.assertEqual(len(self.packet_store._dict), 4)
        self.assertEqual(len(self.packet_store._dict[1]), 2)
        self.assertEqual(len(self.packet_store._dict[0]), 1)
        self.assertEqual(len(self.packet_store._dict[3]), 1)
        self.assertEqual(len(self.packet_store._dict[5]), 1)
        
    def test_get_clse(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")
        self.packet_store.put(arg0=0, arg1=1, cmd=constants.CLSE, data=b"data2")
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd3", data=b"data3")

        self.assertTrue((0, 1) in self.packet_store)
        cmd1, arg0, arg1, data1 = self.packet_store.get(arg0=0, arg1=1)
        self.assertEqual(arg0, 0)
        self.assertEqual(arg1, 1)
        self.assertEqual(cmd1, b"cmd1")
        self.assertEqual(data1, b"data1")
        self.assertTrue((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 1)

        self.assertTrue((0, 1) in self.packet_store)
        cmd2, arg0, arg1, data2 = self.packet_store.get(arg0=0, arg1=1)
        self.assertEqual(arg0, 0)
        self.assertEqual(arg1, 1)
        self.assertEqual(cmd2, constants.CLSE)
        self.assertEqual(data2, b"data2")
        self.assertFalse((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 0)

        self.assertEqual(len(self.packet_store._dict), 0)

    def test_clear(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")
        self.packet_store.put(arg0=2, arg1=1, cmd=b"cmd2", data=b"data2")

        self.packet_store.clear(arg0=None, arg1=None)
        self.assertEqual(len(self.packet_store), 2)

        self.packet_store.clear(arg0=1, arg1=0)
        self.assertEqual(len(self.packet_store), 2)

        self.packet_store.clear(arg0=0, arg1=1)
        self.assertEqual(len(self.packet_store), 1)
        self.assertEqual(len(self.packet_store._dict), 1)

        self.packet_store.clear(arg0=2, arg1=1)
        self.assertEqual(len(self.packet_store), 0)
        self.assertEqual(len(self.packet_store._dict), 0)

    def test_clear_all(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd1", data=b"data1")

        self.packet_store.clear_all()
        self.assertFalse((0, 1) in self.packet_store)
        self.assertEqual(len(self.packet_store), 0)

    def test_find_allow_zeros(self):
        self.packet_store.put(arg0=0, arg1=1, cmd=b"cmd", data=b"data")
        self.assertEqual(self.packet_store.find_allow_zeros(arg0=2, arg1=1), (0, 1))
        self.assertIsNone(self.packet_store.find_allow_zeros(arg0=2, arg1=2))


class TestAdbTransactionInfo(unittest.TestCase):

    def test_args_match(self):
        adb_info_1_None = _AdbTransactionInfo(1, None, 123, 456, 789)
        adb_info_1_2 = _AdbTransactionInfo(1, 2, 123, 456, 789)

        # (1, None) -> exact matches
        self.assertTrue(adb_info_1_None.args_match(6, 1))
        self.assertTrue(adb_info_1_None.args_match(7, 1))
        self.assertTrue(adb_info_1_None.args_match(6, 1, allow_zeros=True))
        self.assertTrue(adb_info_1_None.args_match(7, 1, allow_zeros=True))

        # (1, None) -> no match
        self.assertFalse(adb_info_1_None.args_match(0, 0))
        self.assertFalse(adb_info_1_None.args_match(1, 0))
        self.assertFalse(adb_info_1_None.args_match(2, 0))
        self.assertFalse(adb_info_1_None.args_match(3, 0))
        self.assertFalse(adb_info_1_None.args_match(4, 5, allow_zeros=True))

        # (1, None) -> zero matches
        self.assertTrue(adb_info_1_None.args_match(0, 0, allow_zeros=True))
        self.assertTrue(adb_info_1_None.args_match(1, 0, allow_zeros=True))
        self.assertTrue(adb_info_1_None.args_match(2, 0, allow_zeros=True))
        self.assertTrue(adb_info_1_None.args_match(3, 0, allow_zeros=True))

        # (1, 2) -> exact matches
        self.assertTrue(adb_info_1_2.args_match(2, 1))
        self.assertTrue(adb_info_1_2.args_match(2, 1, allow_zeros=True))

        # (1, 2) -> no match
        self.assertFalse(adb_info_1_2.args_match(0, 0))
        self.assertFalse(adb_info_1_2.args_match(2, 0))
        self.assertFalse(adb_info_1_2.args_match(0, 1))

        self.assertFalse(adb_info_1_2.args_match(1, 2))
        self.assertFalse(adb_info_1_2.args_match(1, 2, allow_zeros=True))

        self.assertFalse(adb_info_1_2.args_match(3, 0))
        self.assertFalse(adb_info_1_2.args_match(0, 4))
        self.assertFalse(adb_info_1_2.args_match(3, 4))
        self.assertFalse(adb_info_1_2.args_match(3, 0, allow_zeros=True))
        self.assertFalse(adb_info_1_2.args_match(0, 4, allow_zeros=True))
        self.assertFalse(adb_info_1_2.args_match(3, 4, allow_zeros=True))

        self.assertFalse(adb_info_1_2.args_match(2, 6))
        self.assertFalse(adb_info_1_2.args_match(2, 6, allow_zeros=True))
        self.assertFalse(adb_info_1_2.args_match(7, 1))
        self.assertFalse(adb_info_1_2.args_match(7, 1, allow_zeros=True))

        # (1, 2) -> zero matches
        self.assertTrue(adb_info_1_2.args_match(0, 0, allow_zeros=True))
        self.assertTrue(adb_info_1_2.args_match(2, 0, allow_zeros=True))
        self.assertTrue(adb_info_1_2.args_match(0, 1, allow_zeros=True))



# Copyright (c) 2021 Jeff Irion and contributors
#
# This file is part of the adb-shell package.  It incorporates work
# covered by the following license notice:
#
#
#   Copyright 2014 Google Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""ADB-related exceptions.

"""


class AdbCommandFailureException(Exception):
    """A ``b'FAIL'`` packet was received.

    """


class AdbConnectionError(Exception):
    """ADB command not sent because a connection to the device has not been established.

    """


class AdbTimeoutError(Exception):
    """ADB command did not complete within the specified time.

    """


class DeviceAuthError(Exception):
    """Device authentication failed.

    """
    def __init__(self, message, *args):
        message %= args
        super(DeviceAuthError, self).__init__(message, *args)


class InvalidChecksumError(Exception):
    """Checksum of data didn't match expected checksum.

    """


class InvalidCommandError(Exception):
    """Got an invalid command.

    """


class InvalidTransportError(Exception):
    """The provided transport does not implement the necessary methods: ``close``, ``connect``, ``bulk_read``, and ``bulk_write``.

    """


class InvalidResponseError(Exception):
    """Got an invalid response to our command.

    """


class DevicePathInvalidError(Exception):
    """A file command was passed an invalid path.

    """


class PushFailedError(Exception):
    """Pushing a file failed for some reason.

    """


class TcpTimeoutException(Exception):
    """TCP connection timed read/write operation exceeded the allowed time.

    """


class UsbDeviceNotFoundError(Exception):
    """TODO

    """


class UsbReadFailedError(Exception):
    """TODO

    Parameters
    ----------
    msg : str
        The error message
    usb_error : libusb1.USBError
        An exception from ``libusb1``

    Attributes
    ----------
    usb_error : libusb1.USBError
        An exception from ``libusb1``

    """
    def __init__(self, msg, usb_error):
        super(UsbReadFailedError, self).__init__(msg, usb_error)
        self.usb_error = usb_error

    def __str__(self):
        return '%s: %s' % self.args


class UsbWriteFailedError(Exception):
    """:meth:`adb_shell.transport.usb_transport.UsbTransport.bulk_write` failed.

    """



import os
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from adb_shell import constants
from adb_shell.adb_device import AdbDevice
from adb_shell.adb_message import AdbMessage, checksum, int_to_cmd, unpack


class TestAdbMessage(unittest.TestCase):
    def test_checksum_bytearray(self):
        cs = checksum(bytearray('TEST', 'utf-8'))
        self.assertEqual(cs, 320)

    def test_checksum_bytes(self):
        cs = checksum(b'TEST')
        self.assertEqual(cs, 320)

    def test_checksum_unicode(self):
        cs = checksum(u'TEST')
        self.assertEqual(cs, 320)

    def test_unpack_error(self):
        with self.assertRaises(ValueError):
            unpack(b'TEST')

    def test_constants(self):
        for key, val in constants.ID_TO_WIRE.items():
            self.assertEqual(key, int_to_cmd(val))



import functools
import inspect
import pickle
import re
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

import adb_shell.exceptions

try:
    getargspec = inspect.getfullargspec
except AttributeError:
    getargspec = inspect.getargspec

try:
    _assertRegex = unittest.TestCase.assertRegex
except AttributeError:
    _assertRegex = unittest.TestCase.assertRegexpMatches


class TestExceptionSerialization(unittest.TestCase):
    def __test_serialize_one_exc_cls(exc_cls):
        # Work out how many args we need to instantiate this object
        try:
            exc_required_arity = len(getargspec(exc_cls.__init__).args)
        except TypeError:
            # In Python 2.7 this could be a slot wrapper which means `__init__`
            # wasn't overridden by the exception subclass - use 0 arity.
            exc_required_arity = 0
        # Don't try to provide `self` - we assume strings will be fine here
        fake_args = ("foo", ) * (exc_required_arity - 1)
        # Instantiate the exception object and then attempt a serializion cycle
        # using `pickle` - we mainly care about whether this blows up or not
        exc_obj = exc_cls(*fake_args)
        pickled_exc_data = pickle.dumps(exc_obj)
        depickled_exc_obj = pickle.loads(pickled_exc_data)

    for __obj in adb_shell.exceptions.__dict__.values():
        if isinstance(__obj, type) and issubclass(__obj, BaseException):
            __test_method = functools.partial(
                __test_serialize_one_exc_cls, __obj
            )
            __test_name = "test_serialize_{}".format(__obj.__name__)
            locals()[__test_name] = __test_method

    # We want to confirm what the stringification and representation of
    # `UsbReadFailedError` look like since it's a non-trivial subclass of
    # `Exception`
    def test_usbreadfailederror_as_str(self):
        exc_args = (mock.sentinel.error_msg, mock.sentinel.usb1_exc_obj)
        exc_obj = adb_shell.exceptions.UsbReadFailedError(*exc_args)
        expected_str = "{}: {}".format(*exc_args)
        _assertRegex(self, str(exc_obj), re.escape(expected_str))

    def test_usbreadfailederror_as_repr(self):
        exc_args = (mock.sentinel.error_msg, mock.sentinel.usb1_exc_obj)
        exc_obj = adb_shell.exceptions.UsbReadFailedError(*exc_args)
        expected_repr = "{}{!r}".format(
            adb_shell.exceptions.UsbReadFailedError.__name__, exc_args
        )
        _assertRegex(self, repr(exc_obj), re.escape(expected_repr))



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ADB authentication using ``pycryptodome``.


.. rubric:: Contents

* :class:`PycryptodomeAuthSigner`

    * :meth:`PycryptodomeAuthSigner.GetPublicKey`
    * :meth:`PycryptodomeAuthSigner.Sign`

"""

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15


class PycryptodomeAuthSigner(object):
    """AuthSigner using the pycryptodome package.

    Parameters
    ----------
    rsa_key_path : str, None
        The path to the private key

    Attributes
    ----------
    public_key : str
        The contents of the public key file
    rsa_key : Crypto.PublicKey.RSA.RsaKey
        The contents of theprivate key

    """
    def __init__(self, rsa_key_path=None):
        super(PycryptodomeAuthSigner, self).__init__()

        if rsa_key_path:
            with open(rsa_key_path + '.pub', 'rb') as rsa_pub_file:
                self.public_key = rsa_pub_file.read()

            with open(rsa_key_path, 'rb') as rsa_priv_file:
                self.rsa_key = RSA.import_key(rsa_priv_file.read())

    def Sign(self, data):
        """Signs given data using a private key.

        Parameters
        ----------
        data : bytes, bytearray
            The data to be signed

        Returns
        -------
        bytes
            The signed ``data``

        """
        h = SHA256.new(data)
        return pkcs1_15.new(self.rsa_key).sign(h)

    def GetPublicKey(self):
        """Returns the public key in PEM format without headers or newlines.

        Returns
        -------
        self.public_key : str
            The contents of the public key file

        """
        return self.public_key



import struct

from adb_shell import constants


class FileSyncMessage(object):  # pylint: disable=too-few-public-methods
    """A helper class for packing FileSync messages.

    Parameters
    ----------
    command : bytes
        TODO
    arg0 : int
        TODO
    data : bytes
        The data that will be sent

    Attributes
    ----------
    arg0 : int
        TODO
    command : int
        The input parameter ``command`` converted to an integer via :const:`adb_shell.constants.FILESYNC_ID_TO_WIRE`
    data : bytes
        The data that will be sent

    """
    def __init__(self, command, arg0=None, data=b''):
        self.command = constants.FILESYNC_ID_TO_WIRE[command]
        self.arg0 = arg0 or len(data)
        self.data = data

    def pack(self):
        """Returns this message in an over-the-wire format.

        Returns
        -------
        bytes
            The message packed into the format required by ADB

        """
        return struct.pack(b'<2I', self.command, self.arg0)


class FileSyncListMessage(object):  # pylint: disable=too-few-public-methods
    """A helper class for packing FileSync messages for the "list" service".

    Parameters
    ----------
    command : bytes
        TODO
    arg0 : int
        TODO
    arg1 : TODO
        TODO
    arg2 : TODO
        TODO
    data : bytes
        The data that will be sent

    Attributes
    ----------
    arg0 : int
        TODO
    arg1 : TODO
        TODO
    arg2 : TODO
        TODO
    arg3 : int
        The size of the data
    command : int
        The input parameter ``command`` converted to an integer via :const:`adb_shell.constants.FILESYNC_ID_TO_WIRE`
    data : bytes
        TODO

    """
    def __init__(self, command, arg0, arg1, arg2, data=b''):
        self.command = constants.FILESYNC_ID_TO_WIRE[command]
        self.arg0 = arg0
        self.arg1 = arg1
        self.arg2 = arg2
        self.arg3 = len(data)
        self.data = data

    def pack(self):
        """Returns this message in an over-the-wire format.

        Returns
        -------
        bytes
            The message packed into the format required by ADB

        """
        return struct.pack(b'<5I', self.command, self.arg0, self.arg1, self.arg2, self.arg3)


class FileSyncStatMessage(object):  # pylint: disable=too-few-public-methods
    """A helper class for packing FileSync messages for the "stat" service".

    Parameters
    ----------
    command : bytes
        TODO
    arg0 : int
        TODO
    arg1 : TODO
        TODO
    arg2 : TODO
        TODO

    Attributes
    ----------
    arg0 : int
        TODO
    arg1 : TODO
        TODO
    arg2 : TODO
        TODO
    command : int
        The input parameter ``command`` converted to an integer via :const:`adb_shell.constants.FILESYNC_ID_TO_WIRE`
    data : bytes
        The data that will be sent (always empty)

    """
    def __init__(self, command, arg0, arg1, arg2):
        self.command = constants.FILESYNC_ID_TO_WIRE[command]
        self.arg0 = arg0
        self.arg1 = arg1
        self.arg2 = arg2
        self.data = b''

    def pack(self):
        """Returns this message in an over-the-wire format.

        Returns
        -------
        bytes
            The message packed into the format required by ADB

        """
        return struct.pack(b'<4I', self.command, self.arg0, self.arg1, self.arg2)



import inspect
import logging
from io import BytesIO
import struct
import sys
import time
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from adb_shell import adb_device, constants, exceptions
from adb_shell.adb_device import AdbDevice, AdbDeviceTcp, DeviceFile
from adb_shell.adb_message import AdbMessage
from adb_shell.auth.keygen import keygen
from adb_shell.auth.sign_pythonrsa import PythonRSASigner

from . import patchers
from .filesync_helpers import FileSyncMessage, FileSyncListMessage, FileSyncStatMessage
from .keygen_stub import open_priv_pub


# https://stackoverflow.com/a/7483862
_LOGGER = logging.getLogger('adb_shell.adb_device')
_LOGGER.setLevel(logging.DEBUG)
_LOGGER.addHandler(logging.StreamHandler(sys.stdout))


def to_int(cmd):
    return sum(c << (i * 8) for i, c in enumerate(bytearray(cmd)))

def join_messages(*messages):
    return b''.join([message.pack() + message.data for message in messages])


class AdbMessageForTesting(AdbMessage):
    def __init__(self, command, arg0=None, arg1=None, data=b''):
        self.command = to_int(command)
        self.magic = self.command ^ 0xFFFFFFFF
        self.arg0 = arg0
        self.arg1 = arg1
        self.data = data


class TestAdbDevice(unittest.TestCase):
    def setUp(self):
        self.transport = patchers.FakeTcpTransport('host', 5555)
        self.device = AdbDevice(transport=self.transport)
        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST)
        self.progress_callback_count = 0

        def _progress_callback(device_path, current, total_bytes):
            print("device_path = {}, current = {}, total_bytes = {}".format(device_path, current, total_bytes))
            self.progress_callback_count += 1

        self.progress_callback = _progress_callback

    def tearDown(self):
        self.assertFalse(self.transport.bulk_read_data)
        self.assertEqual(len(self.device._io_manager._packet_store._dict), 0)

    @staticmethod
    def fake_stat(*args, **kwargs):
        return 1, 2, 3

    def test_no_async_references(self):
        """Make sure there are no references to async code."""
        adb_device_source = inspect.getsource(adb_device)
        self.assertTrue("base_transport_async" not in adb_device_source)
        self.assertTrue("BaseTransportAsync" not in adb_device_source)
        self.assertTrue("adb_device_async" not in adb_device_source)
        self.assertTrue("AdbDeviceAsync" not in adb_device_source)
        self.assertTrue("async" not in adb_device_source)
        self.assertTrue("Async" not in adb_device_source)
        self.transport.bulk_read_data = b''

    def test_adb_connection_error(self):
        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.exec_out('FAIL')

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.root()

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.shell('FAIL')

        with self.assertRaises(exceptions.AdbConnectionError):
            ''.join(self.device.streaming_shell('FAIL'))

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.reboot()

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.root()

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.list('FAIL')

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.push('FAIL', 'FAIL')

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.pull('FAIL', 'FAIL')

        with self.assertRaises(exceptions.AdbConnectionError):
            self.device.stat('FAIL')

        self.transport.bulk_read_data = b''

    def test_init_tcp(self):
        with patchers.PATCH_TCP_TRANSPORT:
            tcp_device = AdbDeviceTcp('host')
            tcp_device._io_manager._transport.bulk_read_data = self.transport.bulk_read_data

        # Make sure that the `connect()` method works
        self.assertTrue(tcp_device.connect())
        self.assertTrue(tcp_device.available)

        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''
        
    def test_init_banner(self):
        device_with_banner = AdbDevice(transport=patchers.FakeTcpTransport('host', 5555), banner='banner')
        self.assertEqual(device_with_banner._banner, b'banner')

        device_with_banner2 = AdbDevice(transport=patchers.FakeTcpTransport('host', 5555), banner=bytearray('banner2', 'utf-8'))
        self.assertEqual(device_with_banner2._banner, b'banner2')

        device_with_banner3 = AdbDevice(transport=patchers.FakeTcpTransport('host', 5555), banner=u'banner3')
        self.assertEqual(device_with_banner3._banner, b'banner3')

        with patch('socket.gethostname', side_effect=Exception):
            device_banner_unknown = AdbDevice(transport=self.transport)
            self.assertTrue(device_banner_unknown.connect())
            self.assertEqual(device_banner_unknown._banner, b'unknown')

    def test_init_invalid_transport(self):
        with self.assertRaises(exceptions.InvalidTransportError):
            device = AdbDevice(transport=123)

        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_available(self):
        self.assertFalse(self.device.available)

        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_close(self):
        self.assertFalse(self.device.close())
        self.assertFalse(self.device.available)

        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    # ======================================================================= #
    #                                                                         #
    #                             `connect` tests                             #
    #                                                                         #
    # ======================================================================= #
    def test_connect(self):
        self.assertTrue(self.device.connect())
        self.assertTrue(self.device.available)

    def test_connect_no_keys(self):
        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST_WITH_AUTH[:2])
        with self.assertRaises(exceptions.DeviceAuthError):
            self.device.connect()

        self.assertFalse(self.device.available)

    def test_connect_with_key_invalid_response(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')

        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST_WITH_AUTH_INVALID)

        with self.assertRaises(exceptions.InvalidResponseError):
            self.device.connect([signer])

        self.assertFalse(self.device.available)

    def test_connect_with_key(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')

        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST_WITH_AUTH)

        self.assertTrue(self.device.connect([signer]))

    def test_connect_with_new_key(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')
            signer.pub_key = u''

        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST_WITH_AUTH_NEW_KEY)

        self.assertTrue(self.device.connect([signer]))

    def test_connect_with_new_key_and_callback(self):
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')
            signer.pub_key = u''

        self._callback_invoked = False
        def auth_callback(device):
            self._callback_invoked = True

        self.transport.bulk_read_data = b''.join(patchers.BULK_READ_LIST_WITH_AUTH_NEW_KEY)

        self.assertTrue(self.device.connect([signer], auth_callback=auth_callback))
        self.assertTrue(self._callback_invoked)

    def test_connect_timeout(self):
        self.transport.bulk_read_data = AdbMessage(command=constants.CLSE, arg0=1, arg1=1).pack()

        with self.assertRaises(exceptions.AdbTimeoutError):
            # Use a negative timeout to ensure that only one packet gets read
            self.device.connect([], read_timeout_s=-1)

    # ======================================================================= #
    #                                                                         #
    #                              `shell` tests                              #
    #                                                                         #
    # ======================================================================= #
    def test_shell_no_return(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(self.device.shell('TEST'), '')

    def test_shell_return_pass(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PA'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'SS'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(self.device.shell('TEST'), 'PASS')

    def test_shell_local_id_wraparound(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=2**32 - 1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=2**32 - 1, data=b'PASS1'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=2**32 - 1, data=b''),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PASS2'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.device._local_id = 2**32 - 2
        self.assertEqual(self.device.shell('TEST'), 'PASS1')
        self.assertEqual(self.device.shell('TEST'), 'PASS2')

    def test_shell_return_pass_with_unexpected_packet(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PA'),
                                                      AdbMessage(command=constants.AUTH, arg0=1, arg1=1, data=b'UNEXPECTED'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'SS'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(self.device.shell('TEST'), 'PASS')

    def test_shell_dont_decode(self):
        self.assertTrue(self.device.connect())
        
        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PA'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'SS'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(self.device.shell('TEST', decode=False), b'PASS')

    def test_shell_avoid_decode_error(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'\x80abc'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        expected = '\\x80abc' if sys.version_info[0] > 2 else u'\ufffdabc'
        self.assertEqual(self.device.shell('TEST'), expected)

    def test_shell_data_length_exceeds_max(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'0'*(constants.MAX_ADB_DATA+1)),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.device.shell('TEST')
        self.assertTrue(True)

    def test_shell_multibytes_sequence_exceeds_max(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'0'*(constants.MAX_ADB_DATA-1) + b'\xe3\x81\x82'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(u'0'*(constants.MAX_ADB_DATA-1) + u'\u3042', self.device.shell('TEST'))

    def test_shell_with_multibytes_sequence_over_two_messages(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'\xe3'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'\x81\x82'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(u'\u3042', self.device.shell('TEST'))

    def test_shell_multiple_clse(self):
        # https://github.com/JeffLIrion/adb_shell/issues/15#issuecomment-536795938
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        msg1 = AdbMessage(command=constants.OKAY, arg0=2, arg1=2, data=b'')
        msg2 = AdbMessage(command=constants.WRTE, arg0=2, arg1=2, data=b'PASS')
        msg3 = AdbMessage(command=constants.CLSE, arg0=2, arg1=2, data=b'')
        self.transport.bulk_read_data = b''.join([b'OKAY\xd9R\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',
                                                  b'WRTE\xd9R\x00\x00\x01\x00\x00\x00\x01\x00\x00\x002\x00\x00\x00\xa8\xad\xab\xba',
                                                  b'2',
                                                  b'WRTE\xd9R\x00\x00\x01\x00\x00\x00\x0c\x02\x00\x00\xc0\x92\x00\x00\xa8\xad\xab\xba',
                                                  b'Wake Locks: size=2\ncom.google.android.tvlauncher\n\n- STREAM_MUSIC:\n   Muted: true\n   Min: 0\n   Max: 15\n   Current: 2 (speaker): 15, 4 (headset): 10, 8 (headphone): 10, 80 (bt_a2dp): 10, 1000 (digital_dock): 10, 4000000 (usb_headset): 3, 40000000 (default): 15\n   Devices: speaker\n- STREAM_ALARM:\n   Muted: true\n   Min: 1\n   Max: 7\n   Current: 2 (speaker): 7, 4 (headset): 5, 8 (headphone): 5, 80 (bt_a2dp): 5, 1000 (digital_dock): 5, 4000000 (usb_headset): 1, 40000000 (default): 7\n   Devices: speaker\n- STREAM_NOTIFICATION:\n',
                                                  b'CLSE\xd9R\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',
                                                  msg1.pack(),
                                                  b'CLSE\xdaR\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',
                                                  msg2.pack(),
                                                  msg2.data,
                                                  msg3.pack()])

        self.device.shell("dumpsys power | grep 'Display Power' | grep -q 'state=ON' && echo -e '1\\c' && dumpsys power | grep mWakefulness | grep -q Awake && echo -e '1\\c' && dumpsys audio | grep paused | grep -qv 'Buffer Queue' && echo -e '1\\c' || (dumpsys audio | grep started | grep -qv 'Buffer Queue' && echo '2\\c' || echo '0\\c') && dumpsys power | grep Locks | grep 'size=' && CURRENT_APP=$(dumpsys window windows | grep mCurrentFocus) && CURRENT_APP=${CURRENT_APP#*{* * } && CURRENT_APP=${CURRENT_APP%%/*} && echo $CURRENT_APP && (dumpsys media_session | grep -A 100 'Sessions Stack' | grep -A 100 $CURRENT_APP | grep -m 1 'state=PlaybackState {' || echo) && dumpsys audio | grep '\\- STREAM_MUSIC:' -A 12")
        self.assertEqual(self.device.shell('TEST'), 'PASS')

    def test_shell_multiple_streams(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=2, data=b'PASS2'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PASS1'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=2, data=b''))

        self.assertEqual(self.device.shell('TEST1'), 'PASS1')
        self.assertEqual(self.device.shell('TEST2'), 'PASS2')

    def test_shell_multiple_streams2(self):
        self.assertTrue(self.device.connect())

        def fake_read_packet_from_device(*args, **kwargs):
            # Mimic the scenario that this stream's packets get read by another stream after the first attempt to read the packet from the device
            self.device._io_manager._packet_store.put(arg0=1, arg1=1, cmd=constants.WRTE, data=b'\x00')
            self.device._io_manager._packet_store.put(arg0=1, arg1=1, cmd=constants.OKAY, data=b'\x00')
            self.device._io_manager._packet_store.put(arg0=2, arg1=2, cmd=constants.OKAY, data=b'\x00')
            self.device._io_manager._packet_store.put(arg0=1, arg1=1, cmd=constants.OKAY, data=b'\x00')
            self.device._io_manager._packet_store.put(arg0=2, arg1=2, cmd=constants.WRTE, data=b'PASS2')
            self.device._io_manager._packet_store.put(arg0=1, arg1=1, cmd=constants.WRTE, data=b"PASS1")
            self.device._io_manager._packet_store.put(arg0=1, arg1=1, cmd=constants.CLSE, data=b"")
            self.device._io_manager._packet_store.put(arg0=2, arg1=2, cmd=constants.CLSE, data=b"")

            return constants.OKAY, 2, 2, b"\x00"

        with patch.object(self.device._io_manager, "_read_packet_from_device", fake_read_packet_from_device):
            # The patch function will only be called once, all subsequent packets will be retrieved from the store
            self.assertEqual(self.device.shell('TEST1'), 'PASS1')
            self.assertEqual(self.device.shell('TEST2'), 'PASS2')

    def test_shell_local_id2(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=2, data=b'PASS2'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PASS1'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=2, data=b''))

        self.assertEqual(self.device.shell('TEST1'), 'PASS1')
        self.assertEqual(self.device.shell('TEST2'), 'PASS2')

    def test_shell_remote_id2(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=2, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=2, arg1=2, data=b'PASS2'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PASS1'),
                                                      AdbMessage(command=constants.CLSE, arg0=2, arg1=2, data=b''),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual(self.device.shell('TEST1'), 'PASS1')
        self.assertEqual(self.device.shell('TEST2'), 'PASS2')

    # ======================================================================= #
    #                                                                         #
    #                           `shell` error tests                           #
    #                                                                         #
    # ======================================================================= #
    def test_shell_error_local_id_timeout(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1234, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1234, data=b'\x00'))

        with self.assertRaises(exceptions.AdbTimeoutError):
            self.device.shell('TEST', read_timeout_s=1)

        # Close the connection so that the packet store gets cleared
        self.device.close()

    def test_shell_error_unknown_command(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessageForTesting(command=constants.FAIL, arg0=1, arg1=1, data=b''))

        with self.assertRaises(exceptions.InvalidCommandError):
            self.assertEqual(self.device.shell('TEST'), '')

    def test_shell_error_transport_timeout(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b''))

        with self.assertRaises(exceptions.AdbTimeoutError):
            self.device.shell('TEST', read_timeout_s=-1)

    def test_shell_error_read_timeout_multiple_clse(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.CLSE, arg0=2, arg1=1, data=b''))

        with self.assertRaises(exceptions.AdbTimeoutError):
            self.device.shell('TEST', read_timeout_s=-1)

    def test_shell_error_timeout(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PA'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'SS'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        def fake_read_until(*args, **kwargs):
            time.sleep(0.2)
            return b'WRTE', b'PA'

        with patch('adb_shell.adb_device.AdbDevice._read_until', fake_read_until):
            with self.assertRaises(exceptions.AdbTimeoutError):
                self.device.shell('TEST', timeout_s=0.5)

        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_shell_error_checksum(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        msg1 = AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00')
        msg2 = AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'PASS')
        self.transport.bulk_read_data = b''.join([msg1.pack(), msg1.data, msg2.pack(), msg2.data[:-1] + b'0'])

        with self.assertRaises(exceptions.InvalidChecksumError):
            self.device.shell('TEST')

    def test_issue29(self):
        # https://github.com/JeffLIrion/adb_shell/issues/29
        with patch('adb_shell.auth.sign_pythonrsa.open', open_priv_pub), patch('adb_shell.auth.keygen.open', open_priv_pub):
            keygen('tests/adbkey')
            signer = PythonRSASigner.FromRSAKeyPath('tests/adbkey')

        okay3 = AdbMessage(command=constants.OKAY, arg0=1, arg1=3, data=b'\x00')
        clse3 = AdbMessage(command=constants.CLSE, arg0=1, arg1=3, data=b'')
        okay5 = AdbMessage(command=constants.OKAY, arg0=1, arg1=5, data=b'\x00')
        clse5 = AdbMessage(command=constants.CLSE, arg0=1, arg1=5, data=b'')
        okay7 = AdbMessage(command=constants.OKAY, arg0=1, arg1=7, data=b'\x00')
        clse7 = AdbMessage(command=constants.CLSE, arg0=1, arg1=7, data=b'')

        self.transport.bulk_read_data = b''.join([b'AUTH\x01\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00\xc5\n\x00\x00\xbe\xaa\xab\xb7',  # Line 22
                                                  b"\x17\xbf\xbf\xff\xc7\xa2eo'Sh\xdf\x8e\xf5\xff\xe0\tJ6H",  # Line 23
                                                  b"CNXN\x00\x00\x00\x01\x00\x10\x00\x00i\x00\x00\x00.'\x00\x00\xbc\xb1\xa7\xb1",  # Line 26
                                                  b'device::ro.product.name=once;ro.product.model=MIBOX3;ro.product.device=once;features=stat_v2,cmd,shell_v2',  # Line 27
                                                  b'OKAY\x99\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',  # Line 290 (modified --> Line 30)
                                                  b'CLSE\xa2\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 291
                                                  b'CLSE\xa2\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 292
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x001\x00\x00\x00\xa8\xad\xab\xba',  # Line 31
                                                  b'1',  # Line 32
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x001\x00\x00\x00\xa8\xad\xab\xba',  # Line 35
                                                  b'1',  # Line 36
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x000\x00\x00\x00\xa8\xad\xab\xba',  # Line 39
                                                  b'0',  # Line 40
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x13\x00\x00\x000\x06\x00\x00\xa8\xad\xab\xba',  # Line 43
                                                  b'Wake Locks: size=0\n',  # Line 44
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x1e\x00\x00\x00V\x0b\x00\x00\xa8\xad\xab\xba',  # Line 47
                                                  b'com.google.android.youtube.tv\n',  # Line 48
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00\x98\x00\x00\x00\xa13\x00\x00\xa8\xad\xab\xba',  # Line 51
                                                  b'      state=PlaybackState {state=0, position=0, buffered position=0, speed=0.0, updated=0, actions=0, custom actions=[], active item id=-1, error=null}\n',  # Line 52
                                                  b'WRTE\x99\x00\x00\x00\x01\x00\x00\x00.\x01\x00\x00\xceP\x00\x00\xa8\xad\xab\xba',  # Line 55
                                                  b'- STREAM_MUSIC:\n   Muted: false\n   Min: 0\n   Max: 15\n   Current: 2 (speaker): 11, 4 (headset): 10, 8 (headphone): 10, 400 (hdmi): 6, 40000000 (default): 11\n   Devices: hdmi\n- STREAM_ALARM:\n   Muted: false\n   Min: 0\n   Max: 7\n   Current: 40000000 (default): 6\n   Devices: speaker\n- STREAM_NOTIFICATION:\n',  # Line 56
                                                  b'CLSE\x99\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 59
                                                  b'AUTH\x01\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00\x94\t\x00\x00\xbe\xaa\xab\xb7',  # Line 297
                                                  b'P\xa5\x86\x97\xe8\x01\xb09\x8c>F\x9d\xc6\xbd\xc0J\x80!\xbb\x1a',  # Line 298
                                                  b"CNXN\x00\x00\x00\x01\x00\x10\x00\x00i\x00\x00\x00.'\x00\x00\xbc\xb1\xa7\xb1",  # Line 301
                                                  b'device::ro.product.name=once;ro.product.model=MIBOX3;ro.product.device=once;features=stat_v2,cmd,shell_v2',  # Line 302
                                                  b'OKAY\xa5\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',  # Line 305
                                                  b'CLSE\xa5\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 306
                                                  okay3.pack(),
                                                  okay3.data,
                                                  clse3.pack(),
                                                  b'AUTH\x01\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00e\x0c\x00\x00\xbe\xaa\xab\xb7',  # Line 315
                                                  b'\xd3\xef\x7f_\xa6\xc0`b\x19\\z\xe4\xf3\xe2\xed\x8d\xe1W\xfbH',  # Line 316
                                                  b"CNXN\x00\x00\x00\x01\x00\x10\x00\x00i\x00\x00\x00.'\x00\x00\xbc\xb1\xa7\xb1",  # Line 319
                                                  b'device::ro.product.name=once;ro.product.model=MIBOX3;ro.product.device=once;features=stat_v2,cmd,shell_v2',  # Line 320
                                                  b'OKAY\xa7\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',  # Line 323
                                                  b'CLSE\xa7\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 324
                                                  okay5.pack(),
                                                  okay5.data,
                                                  clse5.pack(),
                                                  b'AUTH\x01\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00\x93\x08\x00\x00\xbe\xaa\xab\xb7',  # Line 333
                                                  b's\xd4_e\xa4s\x02\x95\x0f\x1e\xec\n\x95Y9[`\x8e\xe1f',  # Line 334
                                                  b"CNXN\x00\x00\x00\x01\x00\x10\x00\x00i\x00\x00\x00.'\x00\x00\xbc\xb1\xa7\xb1",  # Line 337
                                                  b'device::ro.product.name=once;ro.product.model=MIBOX3;ro.product.device=once;features=stat_v2,cmd,shell_v2',  # Line 338
                                                  b'OKAY\xa9\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',  # Line 341
                                                  b'CLSE\xa9\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',  # Line 342
                                                  okay7.pack(),
                                                  okay7.data,
                                                  clse7.pack()])

        self.assertTrue(self.device.connect([signer]))

        self.device.shell('Android TV update command')
        
        self.assertTrue(self.device.connect([signer]))
        self.device.shell('Android TV update command')
        self.device.shell('Android TV update command')
        self.assertTrue(self.device.connect([signer]))
        self.device.shell('Android TV update command')
        self.device.shell('Android TV update command')
        self.assertTrue(self.device.connect([signer]))
        self.device.shell('Android TV update command')
        self.device.shell('Android TV update command')

    # ======================================================================= #
    #                                                                         #
    #                      `streaming_shell` tests                            #
    #                                                                         #
    # ======================================================================= #
    def test_streaming_shell_decode(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(
            AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'ABC'),
            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'123'),
        )

        generator = self.device.streaming_shell('TEST', decode=True)
        self.assertEqual('ABC', next(generator))
        self.assertEqual('123', next(generator))

    def test_streaming_shell_dont_decode(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(
            AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'ABC'),
            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'123'),
        )

        generator = self.device.streaming_shell('TEST', decode=False)
        self.assertEqual(b'ABC', next(generator))
        self.assertEqual(b'123', next(generator))


    # ======================================================================= #
    #                                                                         #
    #                              `reboot` test                              #
    #                                                                         #
    # ======================================================================= #
    def test_reboot(self):
        self.assertTrue(self.device.connect())

        with patch('adb_shell.adb_device.AdbDevice._open') as patch_open:
            self.device.reboot()
            assert patch_open.call_count == 1


    # ======================================================================= #
    #                                                                         #
    #                               `root` test                               #
    #                                                                         #
    # ======================================================================= #
    def test_root(self):
        self.assertTrue(self.device.connect())

        with patch('adb_shell.adb_device.AdbDevice._service') as patch_service:
            self.device.root()
            assert patch_service.call_count == 1


    # ======================================================================= #
    #                                                                         #
    #                          `exec_out` test                                #
    #                                                                         #
    # ======================================================================= #
    def test_exec_out(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = b''.join([b'OKAY\x14\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',
                                                  b'WRTE\x14\x00\x00\x00\x01\x00\x00\x00\x05\x00\x00\x00J\x01\x00\x00\xa8\xad\xab\xba',
                                                  b'TEST\n',
                                                  b'',
                                                  b'CLSE\x14\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba'])

        self.assertEqual(self.device.exec_out("echo 'TEST'"), "TEST\n")

    # ======================================================================= #
    #                                                                         #
    #                         `filesync` tests                                #
    #                                                                         #
    # ======================================================================= #
    def test_list(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncListMessage(constants.DENT, 1, 2, 3, data=b'file1'),
                                                                                                                            FileSyncListMessage(constants.DENT, 4, 5, 6, data=b'file2'),
                                                                                                                            FileSyncListMessage(constants.DONE, 0, 0, 0))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.LIST, data=b'/dir'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        expected_result = [DeviceFile(filename=bytearray(b'file1'), mode=1, size=2, mtime=3),
                           DeviceFile(filename=bytearray(b'file2'), mode=4, size=5, mtime=6)]

        self.assertEqual(expected_result, self.device.list('/dir'))
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_list_empty_path(self):
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.list("")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.list(b"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.list(u"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.list(None)
        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_push_fail(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 100
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(constants.FAIL, data=b''))))

        with self.assertRaises(exceptions.PushFailedError), patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)):
            self.device.push('TEST_FILE', '/data', mtime=mtime)

    def test_push_file(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 100
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=FileSyncMessage(constants.OKAY).pack()),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.SEND, data=b'/data,33272'),
                                                                                                                  FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                  FileSyncMessage(command=constants.DONE, arg0=mtime, data=b''))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)):
            self.assertEqual(self.progress_callback_count, 0)
            with patch("adb_shell.adb_device.os.fstat", return_value=patchers.StSize(12345)):
                self.device.push('TEST_FILE', '/data', mtime=mtime, progress_callback=self.progress_callback)
            self.assertEqual(self.progress_callback_count, 1)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_push_bytesio(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 100
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=FileSyncMessage(constants.OKAY).pack()),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.SEND, data=b'/data,33272'),
                                                                                                                  FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                  FileSyncMessage(command=constants.DONE, arg0=mtime, data=b''))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        stream = BytesIO(filedata)
        self.device.push(stream, '/data', mtime=mtime)
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_push_file_exception(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 100
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=FileSyncMessage(constants.OKAY).pack()),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.SEND, data=b'/data,33272'),
                                                                                                                  FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                  FileSyncMessage(command=constants.DONE, arg0=mtime, data=b''))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)):
            # Set self.progress_callback_count to None so that an exception occurs when self.progress_callback tries to increment it
            self.progress_callback_count = None
            with patch("adb_shell.adb_device.os.fstat", return_value=patchers.StSize(12345)):
                self.device.push('TEST_FILE', '/data', mtime=mtime, progress_callback=self.progress_callback)
            
            self.assertIsNone(self.progress_callback_count)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_push_file_mtime0(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 0
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(constants.OKAY, data=b''))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.SEND, data=b'/data,33272'),
                                                                                                                  FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                  FileSyncMessage(command=constants.DONE, arg0=mtime))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)), patch('time.time', return_value=mtime):
            self.device.push('TEST_FILE', '/data', mtime=mtime)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_push_big_file(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        mtime = 100
        filedata = b'0' * int(3.5 * self.device.max_chunk_size)

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(constants.OKAY))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        mcs0, mcs1, mcs2, mcs3 = 0, self.device.max_chunk_size, 2*self.device.max_chunk_size, 3*self.device.max_chunk_size
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(
                                                FileSyncMessage(command=constants.SEND, data=b'/data,33272'),
                                                FileSyncMessage(command=constants.DATA, data=filedata[mcs0:mcs1]))),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(
                                                FileSyncMessage(command=constants.DATA, data=filedata[mcs1:mcs2]))),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(
                                                FileSyncMessage(command=constants.DATA, data=filedata[mcs2:mcs3]),
                                                FileSyncMessage(command=constants.DATA, data=filedata[mcs3:]),
                                                FileSyncMessage(command=constants.DONE, arg0=mtime))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)):
            self.assertEqual(self.progress_callback_count, 0)
            with patch("adb_shell.adb_device.os.fstat", return_value=patchers.StSize(12345)):
                self.device.push('TEST_FILE', '/data', mtime=mtime, progress_callback=self.progress_callback)
            self.assertEqual(self.progress_callback_count, 4)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_push_dir(self):
        self.assertTrue(self.device.connect())

        mtime = 100
        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.OKAY, arg0=2, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=2, arg1=2, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=2, arg1=2, data=join_messages(FileSyncMessage(constants.OKAY))),
                                                      AdbMessage(command=constants.CLSE, arg0=2, arg1=2, data=b''),
                                                      AdbMessage(command=constants.OKAY, arg0=3, arg1=3, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=3, arg1=3, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=3, arg1=3, data=join_messages(FileSyncMessage(constants.OKAY))),
                                                      AdbMessage(command=constants.CLSE, arg0=3, arg1=3, data=b''))

        # Expected `bulk_write` values
        #TODO

        with patch('adb_shell.adb_device.open', patchers.mock_open(read_data=filedata)), patch('os.path.isdir', lambda x: x == 'TEST_DIR/'), patch('os.listdir', return_value=['TEST_FILE1', 'TEST_FILE2']):
            self.device.push('TEST_DIR/', '/data', mtime=mtime)

    def test_push_empty_path(self):
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.push("NOTHING", "")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.push("NOTHING", b"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.push("NOTHING", u"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.push("NOTHING", None)
        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_pull_file(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                            FileSyncMessage(command=constants.DONE))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/data'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open()) as m:
            self.assertEqual(self.progress_callback_count, 0)
            with patch("adb_shell.adb_device.AdbDevice.stat", self.fake_stat):
                self.device.pull('/data', 'TEST_FILE', progress_callback=self.progress_callback)

            self.assertEqual(self.progress_callback_count, 1)
            self.assertEqual(m.written, filedata)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_pull_bytesio(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                            FileSyncMessage(command=constants.DONE))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/data'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        stream = BytesIO()
        self.device.pull('/data', stream)
        
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)
        self.assertEqual(stream.getvalue(), filedata)    

    def test_pull_file_exception(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        filedata = b'Ohayou sekai.\nGood morning world!'

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                            FileSyncMessage(command=constants.DONE))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/data'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open()) as m:
            # Set self.progress_callback_count to None so that an exception occurs when self.progress_callback tries to increment it
            self.progress_callback_count = None
            with patch("adb_shell.adb_device.AdbDevice.stat", self.fake_stat):
                self.device.pull('/data', 'TEST_FILE', progress_callback=self.progress_callback)

            self.assertIsNone(self.progress_callback_count)
            self.assertEqual(m.written, filedata)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_pull_big_file(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        filedata = b'0' * int(1.5 * constants.MAX_ADB_DATA)

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.DATA, data=filedata),
                                                                                                                            FileSyncMessage(command=constants.DONE))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/data'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        with patch('adb_shell.adb_device.open', patchers.mock_open()) as m:
            self.assertEqual(self.progress_callback_count, 0)
            with patch("adb_shell.adb_device.AdbDevice.stat", self.fake_stat):
                self.device.pull('/data', 'TEST_FILE', progress_callback=self.progress_callback)
            
            self.assertEqual(self.progress_callback_count, 1)
            self.assertEqual(m.written, filedata)
            self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_pull_empty_path(self):
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.pull("", "NOWHERE")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.pull(b"", "NOWHERE")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.pull(u"", "NOWHERE")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.pull(None, "NOWHERE")
        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_pull_non_existant_path(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'FAIL&\x00\x00\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'open failed: No such file or directory'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/does/not/exist'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))
        with self.assertRaises(exceptions.AdbCommandFailureException):
            self.device.pull("/does/not/exist", "NOWHERE")
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_pull_non_existant_path_2(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=b'FAIL&\x00\x00\x00open failed: No such file or directory'),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.RECV, data=b'/does/not/exist'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))
        with self.assertRaises(exceptions.AdbCommandFailureException):
            self.device.pull("/does/not/exist", "NOWHERE")
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_stat(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncStatMessage(constants.STAT, 1, 2, 3),
                                                                                                                            FileSyncStatMessage(constants.DONE, 0, 0, 0))),
                                                      AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        # Expected `bulk_write` values
        expected_bulk_write = join_messages(AdbMessage(command=constants.OPEN, arg0=1, arg1=0, data=b'sync:\x00'),
                                            AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncMessage(command=constants.STAT, data=b'/data'))),
                                            AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b''),
                                            AdbMessage(command=constants.CLSE, arg0=1, arg1=1, data=b''))

        self.assertEqual((1, 2, 3), self.device.stat('/data'))
        self.assertEqual(expected_bulk_write, self.transport.bulk_write_data)

    def test_stat_empty_path(self):
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.stat("")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.stat(b"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.stat(u"")
        with self.assertRaises(exceptions.DevicePathInvalidError):
            self.device.stat(None)
        # Clear the `_bulk_read` buffer so that `self.tearDown()` passes
        self.transport.bulk_read_data = b''

    def test_stat_issue155(self):
        self.assertTrue(self.device.connect())

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = b"".join([b'CLSE\n\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba',
                                                  b'OKAY\x0b\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',
                                                  b'OKAY\x0b\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\xb4\xbe\xa6',
                                                  b'WRTE\x0b\x00\x00\x00\x01\x00\x00\x00\x10\x00\x00\x00\x96\x04\x00\x00\xa8\xad\xab\xba',
                                                  b'STAT\xedA\x00\x00\x00\x10\x00\x00\xf0\x88[I',
                                                  b'CLSE\x0b\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\xb3\xac\xba'])

        # This is where the expected values come from
        mode = 16877
        size = 4096
        mtime = 1230735600
        self.assertEqual(FileSyncStatMessage(constants.STAT, mode, size, mtime).pack(), b'STAT\xedA\x00\x00\x00\x10\x00\x00\xf0\x88[I')

        self.assertEqual((mode, size, mtime), self.device.stat('/'))

    # ======================================================================= #
    #                                                                         #
    #                  `filesync` hidden methods tests                        #
    #                                                                         #
    # ======================================================================= #
    def test_filesync_read_adb_command_failure_exceptions(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncStatMessage(constants.FAIL, 1, 2, 3),
                                                                                                                            FileSyncStatMessage(constants.DONE, 0, 0, 0))))

        with self.assertRaises(exceptions.AdbCommandFailureException):
            self.device.stat('/data')

    def test_filesync_read_invalid_response_error(self):
        self.assertTrue(self.device.connect())
        self.transport.bulk_write_data = b''

        # Provide the `bulk_read` return values
        self.transport.bulk_read_data = join_messages(AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.OKAY, arg0=1, arg1=1, data=b'\x00'),
                                                      AdbMessage(command=constants.WRTE, arg0=1, arg1=1, data=join_messages(FileSyncStatMessage(constants.DENT, 1, 2, 3),
                                                                                                                            FileSyncStatMessage(constants.DONE, 0, 0, 0))))

        with self.assertRaises(exceptions.InvalidResponseError):
            self.device.stat('/data')



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ADB authentication using the ``cryptography`` package.


.. rubric:: Contents

* :class:`CryptographySigner`

    * :meth:`CryptographySigner.GetPublicKey`
    * :meth:`CryptographySigner.Sign`

"""

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import utils


# pylint: disable=abstract-method
class CryptographySigner(object):
    """AuthSigner using cryptography.io.

    Parameters
    ----------
    rsa_key_path : str
        The path to the private key.

    Attributes
    ----------
    public_key : str
        The contents of the public key file
    rsa_key : cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey
        The loaded private key

    """
    def __init__(self, rsa_key_path):
        with open(rsa_key_path + '.pub', 'rb') as rsa_pub_file:
            self.public_key = rsa_pub_file.read()

        with open(rsa_key_path, 'rb') as rsa_prv_file:
            self.rsa_key = serialization.load_pem_private_key(rsa_prv_file.read(), None, default_backend())

    def Sign(self, data):
        """Signs given data using a private key.

        Parameters
        ----------
        data : TODO
            TODO

        Returns
        -------
        TODO
            The signed ``data``

        """
        return self.rsa_key.sign(data, padding.PKCS1v15(), utils.Prehashed(hashes.SHA1()))

    def GetPublicKey(self):
        """Returns the public key in PEM format without headers or newlines.

        Returns
        -------
        self.public_key : str
            The contents of the public key file

        """
        return self.public_key



# Copyright (c) 2021 Jeff Irion and contributors
#
# This file is part of the adb-shell package.  It incorporates work
# covered by the following license notice:
#
#
#   Copyright 2014 Google Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""A class for creating a USB connection with the device and sending and receiving data.

.. warning::

   USB support is an experimental feature.


* :func:`get_interface`
* :func:`interface_matcher`
* :class:`UsbTransport`

    * :meth:`UsbTransport._find`
    * :meth:`UsbTransport._find_and_open`
    * :meth:`UsbTransport._find_devices`
    * :meth:`UsbTransport._find_first`
    * :meth:`UsbTransport._flush_buffers`
    * :meth:`UsbTransport._open`
    * :meth:`UsbTransport._port_path_matcher`
    * :meth:`UsbTransport._serial_matcher`
    * :meth:`UsbTransport._timeout`
    * :meth:`UsbTransport.bulk_read`
    * :meth:`UsbTransport.bulk_write`
    * :meth:`UsbTransport.close`
    * :meth:`UsbTransport.connect`
    * :attr:`UsbTransport.port_path`
    * :attr:`UsbTransport.serial_number`
    * :attr:`UsbTransport.usb_info`

"""


import logging
import platform
import re
import threading
import warnings
import weakref

import usb1

from .base_transport import BaseTransport

from .. import exceptions


#: Default timeout
DEFAULT_TIMEOUT_S = 10

SYSFS_PORT_SPLIT_RE = re.compile("[,/:.-]")

_LOGGER = logging.getLogger(__name__)

CLASS = usb1.CLASS_VENDOR_SPEC  # pylint: disable=no-member
SUBCLASS = 0x42
PROTOCOL = 0x01


def get_interface(setting):  # pragma: no cover
    """Get the class, subclass, and protocol for the given USB setting.

    Parameters
    ----------
    setting : TODO
        TODO

    Returns
    -------
    TODO
        TODO
    TODO
        TODO
    TODO
        TODO

    """
    return (setting.getClass(), setting.getSubClass(), setting.getProtocol())


def interface_matcher(clazz, subclass, protocol):   # pragma: no cover
    """Returns a matcher that returns the setting with the given interface.

    Parameters
    ----------
    clazz : TODO
        TODO
    subclass : TODO
        TODO
    protocol : TODO
        TODO

    Returns
    -------
    matcher : function
        TODO

    """
    interface = (clazz, subclass, protocol)

    def matcher(device):
        """TODO

        Parameters
        ----------
        device : TODO
            TODO

        Returns
        -------
        TODO, None
            TODO

        """
        for setting in device.iterSettings():
            if get_interface(setting) == interface:
                return setting
        return None

    return matcher


class UsbTransport(BaseTransport):   # pragma: no cover
    """USB communication object. Not thread-safe.

    Handles reading and writing over USB with the proper endpoints, exceptions,
    and interface claiming.

    Parameters
    ----------
    device : usb1.USBDevice
        libusb_device to connect to.
    setting : usb1.USBInterfaceSetting
        libusb setting with the correct endpoints to communicate with.
    usb_info : TODO, None
        String describing the usb path/serial/device, for debugging.
    default_transport_timeout_s : TODO, None
        Timeout in seconds for all I/O.

    Attributes
    ----------
    _default_transport_timeout_s : TODO, None
        Timeout in seconds for all I/O.
    _device : TODO
        libusb_device to connect to.
    _transport : TODO
        TODO
    _interface_number : TODO
        TODO
    _max_read_packet_len : TODO
        TODO
    _read_endpoint : TODO
        TODO
    _setting : TODO
        libusb setting with the correct endpoints to communicate with.
    _usb_info : TODO
        String describing the usb path/serial/device, for debugging.
    _write_endpoint : TODO, None
        TODO

    """
    # We maintain an idempotent `usb1` context object to ensure that device
    # objects we hand back to callers can be used while this class exists
    USB1_CTX = usb1.USBContext()
    USB1_CTX.open()

    _HANDLE_CACHE = weakref.WeakValueDictionary()
    _HANDLE_CACHE_LOCK = threading.Lock()

    def __init__(self, device, setting, usb_info=None, default_transport_timeout_s=None):
        self._setting = setting
        self._device = device
        self._transport = None

        self._interface_number = None
        self._read_endpoint = None
        self._write_endpoint = None

        self._usb_info = usb_info or ''
        self._default_transport_timeout_s = default_transport_timeout_s if default_transport_timeout_s is not None else DEFAULT_TIMEOUT_S
        self._max_read_packet_len = 0

    def close(self):
        """Close the USB connection.

        """
        if self._transport is None:
            return
        try:
            self._transport.releaseInterface(self._interface_number)
            self._transport.close()
        except usb1.USBError:
            _LOGGER.info('USBError while closing transport %s: ', self.usb_info, exc_info=True)
        finally:
            self._transport = None

    def connect(self, transport_timeout_s=None):
        """Create a USB connection to the device.

        Parameters
        ----------
        transport_timeout_s : float, None
            Set the timeout on the USB instance

        """
        read_endpoint = None
        write_endpoint = None

        for endpoint in self._setting.iterEndpoints():
            address = endpoint.getAddress()
            if address & usb1.ENDPOINT_DIR_MASK:  # pylint: disable=no-member
                read_endpoint = address
                # max_read_packet_len = endpoint.getMaxPacketSize()
            else:
                write_endpoint = address

        assert read_endpoint is not None
        assert write_endpoint is not None

        transport = self._device.open()
        iface_number = self._setting.getNumber()
        try:
            if (platform.system() != 'Windows' and transport.kernelDriverActive(iface_number)):
                transport.detachKernelDriver(iface_number)
        except usb1.USBErrorNotFound:  # pylint: disable=no-member
            warnings.warn('Kernel driver not found for interface: %s.', iface_number)

        # # When this object is deleted, make sure it's closed.
        # weakref.ref(self, self.close)

        self._transport = transport
        self._read_endpoint = read_endpoint
        self._write_endpoint = write_endpoint
        self._interface_number = iface_number

        self._transport.claimInterface(self._interface_number)

    def bulk_read(self, numbytes, transport_timeout_s=None):
        """Receive data from the USB device.

        Parameters
        ----------
        numbytes : int
            The maximum amount of data to be received
        transport_timeout_s : float, None
            When the timeout argument is omitted, ``select.select`` blocks until at least one file descriptor is ready. A time-out value of zero specifies a poll and never blocks.

        Returns
        -------
        bytes
            The received data

        Raises
        ------
        adb_shell.exceptions.UsbReadFailedError
            Could not receive data

        """
        if self._transport is None:
            raise exceptions.UsbReadFailedError('This transport has been closed, probably due to another being opened.', None)
        try:
            # python-libusb1 > 1.6 exposes bytearray()s now instead of bytes/str.
            # To support older and newer versions, we ensure everything's bytearray()
            # from here on out.
            return bytes(self._transport.bulkRead(self._read_endpoint, numbytes, timeout=self._timeout_ms(transport_timeout_s)))
        except usb1.USBError as e:
            raise exceptions.UsbReadFailedError('Could not receive data from %s (timeout %sms)' % (self.usb_info, self._timeout_ms(transport_timeout_s)), e)

    def bulk_write(self, data, transport_timeout_s=None):
        """Send data to the USB device.

        Parameters
        ----------
        data : bytes
            The data to be sent
        transport_timeout_s : float, None
            When the timeout argument is omitted, ``select.select`` blocks until at least one file descriptor is ready. A time-out value of zero specifies a poll and never blocks.

        Returns
        -------
        int
            The number of bytes sent

        Raises
        ------
        adb_shell.exceptions.UsbWriteFailedError
            This transport has been closed, probably due to another being opened
        adb_shell.exceptions.UsbWriteFailedError
            Could not send data

        """
        if self._transport is None:
            raise exceptions.UsbWriteFailedError('This transport has been closed, probably due to another being opened.', None)

        try:
            return self._transport.bulkWrite(self._write_endpoint, data, timeout=self._timeout_ms(transport_timeout_s))

        except usb1.USBError as e:
            raise exceptions.UsbWriteFailedError('Could not send data to %s (timeout %sms)' % (self.usb_info, self._timeout_ms(transport_timeout_s)), e)

    def _open(self):
        """Opens the USB device for this setting, and claims the interface.

        """
        # Make sure we close any previous transport open to this usb device.
        port_path = tuple(self.port_path)
        with self._HANDLE_CACHE_LOCK:
            old_transport = self._HANDLE_CACHE.get(port_path)
            if old_transport is not None:
                old_transport.Close()

        self._read_endpoint = None
        self._write_endpoint = None

        for endpoint in self._setting.iterEndpoints():
            address = endpoint.getAddress()
            if address & usb1.USB_ENDPOINT_DIR_MASK:  # pylint: disable=no-member
                self._read_endpoint = address
                self._max_read_packet_len = endpoint.getMaxPacketSize()
            else:
                self._write_endpoint = address

        assert self._read_endpoint is not None
        assert self._write_endpoint is not None

        transport = self._device.open()
        iface_number = self._setting.getNumber()
        try:
            if (platform.system() != 'Windows' and transport.kernelDriverActive(iface_number)):
                transport.detachKernelDriver(iface_number)
        except usb1.USBErrorNotFound:  # pylint: disable=no-member
            warnings.warn('Kernel driver not found for interface: %s.', iface_number)
        transport.claimInterface(iface_number)
        self._transport = transport
        self._interface_number = iface_number

        with self._HANDLE_CACHE_LOCK:
            self._HANDLE_CACHE[port_path] = self
        # When this object is deleted, make sure it's closed.
        weakref.ref(self, self.close)

    def _timeout_ms(self, transport_timeout_s):
        """TODO

        Returns
        -------
        TODO
            TODO

        """
        return int(transport_timeout_s * 1000 if transport_timeout_s is not None else self._default_transport_timeout_s * 1000)

    def _flush_buffers(self):
        """TODO

        Raises
        ------
        adb_shell.exceptions.UsbReadFailedError
            TODO

        """
        while True:
            try:
                self.bulk_read(self._max_read_packet_len, transport_timeout_s=10)
            except exceptions.UsbReadFailedError as e:
                if isinstance(e.usb_error, usb1.USBErrorTimeout):  # pylint: disable=no-member
                    break
                raise

    # ======================================================================= #
    #                                                                         #
    #                               Properties                                #
    #                                                                         #
    # ======================================================================= #
    @property
    def port_path(self):
        """TODO

        Returns
        -------
        TODO
            TODO

        """
        return [self._device.getBusNumber()] + self._device.getPortNumberList()

    @property
    def serial_number(self):
        """TODO

        Returns
        -------
        TODO
            TODO

        """
        return self._device.getSerialNumber()

    @property
    def usb_info(self):
        """TODO

        Returns
        -------
        TODO
            TODO

        """
        try:
            sn = self.serial_number
        except usb1.USBError:
            sn = ''
        if sn and sn != self._usb_info:
            return '%s %s' % (self._usb_info, sn)
        return self._usb_info

    # ======================================================================= #
    #                                                                         #
    #                                Matchers                                 #
    #                                                                         #
    # ======================================================================= #
    @classmethod
    def _port_path_matcher(cls, port_path):
        """Returns a device matcher for the given port path.

        Parameters
        ----------
        port_path : TODO
            TODO

        Returns
        -------
        function
            TODO

        """
        if isinstance(port_path, str):
            # Convert from sysfs path to port_path.
            port_path = [int(part) for part in SYSFS_PORT_SPLIT_RE.split(port_path)]
        return lambda device: device.port_path == port_path

    @classmethod
    def _serial_matcher(cls, serial):
        """Returns a device matcher for the given serial.

        Parameters
        ----------
        serial : TODO
            TODO

        Returns
        -------
        function
            TODO

        """
        return lambda device: device.serial_number == serial

    # ======================================================================= #
    #                                                                         #
    #                                 Finders                                 #
    #                                                                         #
    # ======================================================================= #
    @classmethod
    def _find(cls, setting_matcher, port_path=None, serial=None, default_transport_timeout_s=None):
        """Gets the first device that matches according to the keyword args.

        Parameters
        ----------
        setting_matcher : TODO
            TODO
        port_path : TODO, None
            TODO
        serial : TODO, None
            TODO
        default_transport_timeout_s : TODO, None
            TODO

        Returns
        -------
        TODO
            TODO

        """
        if port_path:
            device_matcher = cls._port_path_matcher(port_path)
            usb_info = port_path
        elif serial:
            device_matcher = cls._serial_matcher(serial)
            usb_info = serial
        else:
            device_matcher = None
            usb_info = 'first'
        return cls._find_first(setting_matcher, device_matcher, usb_info=usb_info, default_transport_timeout_s=default_transport_timeout_s)

    @classmethod
    def _find_and_open(cls, setting_matcher, port_path=None, serial=None, default_transport_timeout_s=None):
        """TODO

        Parameters
        ----------
        setting_matcher : TODO
            TODO
        port_path : TODO, None
            TODO
        serial : TODO, None
            TODO
        default_transport_timeout_s : TODO, None
            TODO

        Returns
        -------
        dev : TODO
            TODO

        """
        dev = cls._find(setting_matcher, port_path=port_path, serial=serial, default_transport_timeout_s=default_transport_timeout_s)
        dev._open()  # pylint: disable=protected-access
        dev._flush_buffers()  # pylint: disable=protected-access
        return dev

    @classmethod
    def _find_devices(cls, setting_matcher, device_matcher=None, usb_info='', default_transport_timeout_s=None):
        """_find and yield the devices that match.

        Parameters
        ----------
        setting_matcher : TODO
            Function that returns the setting to use given a ``usb1.USBDevice``, or ``None``
            if the device doesn't have a valid setting.
        device_matcher : TODO, None
            Function that returns ``True`` if the given ``UsbTransport`` is
            valid. ``None`` to match any device.
        usb_info : str
            Info string describing device(s).
        default_transport_timeout_s : TODO, None
            Default timeout of commands in seconds.

        Yields
        ------
        TODO
            UsbTransport instances

        """
        for device in cls.USB1_CTX.getDeviceIterator(skip_on_error=True):
            setting = setting_matcher(device)
            if setting is None:
                continue

            transport = cls(device, setting, usb_info=usb_info, default_transport_timeout_s=default_transport_timeout_s)
            if device_matcher is None or device_matcher(transport):
                yield transport

    @classmethod
    def _find_first(cls, setting_matcher, device_matcher=None, usb_info='', default_transport_timeout_s=None):
        """Find and return the first matching device.

        Parameters
        ----------
        setting_matcher : TODO
            Function that returns the setting to use given a ``usb1.USBDevice``, or ``None``
            if the device doesn't have a valid setting.
        device_matcher : TODO
            Function that returns ``True`` if the given ``UsbTransport`` is
            valid. ``None`` to match any device.
        usb_info : str
            Info string describing device(s).
        default_transport_timeout_s : TODO, None
            Default timeout of commands in seconds.

        Returns
        -------
        TODO
            An instance of `UsbTransport`

        Raises
        ------
        adb_shell.exceptions.DeviceNotFoundError
            Raised if the device is not available.

        """
        try:
            return next(cls._find_devices(setting_matcher, device_matcher=device_matcher, usb_info=usb_info, default_transport_timeout_s=default_transport_timeout_s))
        except StopIteration:
            raise exceptions.UsbDeviceNotFoundError('No device available, or it is in the wrong configuration.')

    @classmethod
    def find_adb(cls, serial=None, port_path=None, default_transport_timeout_s=None):
        """TODO

        Parameters
        ----------
        serial : TODO
            TODO
        port_path : TODO
            TODO
        default_transport_timeout_s : TODO, None
            Default timeout of commands in seconds.

        Returns
        -------
        UsbTransport
            TODO

        """
        return cls._find(
            interface_matcher(CLASS, SUBCLASS, PROTOCOL),
            serial=serial,
            port_path=port_path,
            default_transport_timeout_s=default_transport_timeout_s
        )

    @classmethod
    def find_all_adb_devices(cls, default_transport_timeout_s=None):
        """Find all ADB devices attached via USB.

        Parameters
        ----------
        default_transport_timeout_s : TODO, None
            Default timeout of commands in seconds.

        Returns
        -------
        generator
            A generator which yields each ADB device attached via USB.

        """
        for dev in cls._find_devices(interface_matcher(CLASS, SUBCLASS, PROTOCOL), default_transport_timeout_s=default_transport_timeout_s):
            yield dev



# Copyright (c) 2021 Jeff Irion and contributors
#
# This file is part of the adb-shell package.

"""A base class for transports used to communicate with a device.

* :class:`BaseTransport`

    * :meth:`BaseTransport.bulk_read`
    * :meth:`BaseTransport.bulk_write`
    * :meth:`BaseTransport.close`
    * :meth:`BaseTransport.connect`

"""


try:
    from abc import ABC, abstractmethod
except ImportError:  # pragma: no cover
    from abc import ABCMeta, abstractmethod

    class ABC(object):  # pylint: disable=too-few-public-methods
        """A Python2-compatible `ABC` class.

        """
        __metaclass__ = ABCMeta


class BaseTransport(ABC):
    """A base transport class.

    """

    @abstractmethod
    def close(self):
        """Close the connection.

        """

    @abstractmethod
    def connect(self, transport_timeout_s):
        """Create a connection to the device.

        Parameters
        ----------
        transport_timeout_s : float, None
            A connection timeout

        """

    @abstractmethod
    def bulk_read(self, numbytes, transport_timeout_s):
        """Read data from the device.

        Parameters
        ----------
        numbytes : int
            The maximum amount of data to be received
        transport_timeout_s : float, None
            A timeout for the read operation

        Returns
        -------
        bytes
            The received data

        """

    @abstractmethod
    def bulk_write(self, data, transport_timeout_s):
        """Send data to the device.

        Parameters
        ----------
        data : bytes
            The data to be sent
        transport_timeout_s : float, None
            A timeout for the write operation

        Returns
        -------
        int
            The number of bytes sent

        """



import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from adb_shell.exceptions import TcpTimeoutException
from adb_shell.transport.tcp_transport import TcpTransport

from . import patchers


class TestTcpTransport(unittest.TestCase):
    def setUp(self):
        """Create a ``TcpTransport`` and connect to a TCP service.

        """
        self.transport = TcpTransport('host', 5555)
        with patchers.PATCH_CREATE_CONNECTION:
            self.transport.connect(transport_timeout_s=1)

    def tearDown(self):
        """Close the socket connection."""
        self.transport.close()

    def test_connect_with_timeout(self):
        """TODO

        """
        self.transport.close()
        with patchers.PATCH_CREATE_CONNECTION:
            self.transport.connect(transport_timeout_s=1)
            self.assertTrue(True)

    def test_bulk_read(self):
        """TODO

        """
        # Provide the `recv` return values
        self.transport._connection._recv = b'TEST1TEST2'

        with patchers.PATCH_SELECT_SUCCESS:
            self.assertEqual(self.transport.bulk_read(5, transport_timeout_s=1), b'TEST1')
            self.assertEqual(self.transport.bulk_read(5, transport_timeout_s=1), b'TEST2')

        with patchers.PATCH_SELECT_FAIL:
            with self.assertRaises(TcpTimeoutException):
                self.transport.bulk_read(4, transport_timeout_s=1)

    def test_close_oserror(self):
        """Test that an `OSError` exception is handled when closing the socket.

        """
        with patch('{}.patchers.FakeSocket.shutdown'.format(__name__), side_effect=OSError):
            self.transport.close()

    def test_bulk_write(self):
        """TODO

        """
        with patchers.PATCH_SELECT_SUCCESS:
            self.transport.bulk_write(b'TEST', transport_timeout_s=1)

        with patchers.PATCH_SELECT_FAIL:
            with self.assertRaises(TcpTimeoutException):
                self.transport.bulk_write(b'FAIL', transport_timeout_s=1)



import asyncio
import warnings



def _await(coro):
    """Create a new event loop, run the coroutine, then close the event loop."""
    loop = asyncio.new_event_loop()

    with warnings.catch_warnings(record=True) as warns:
        ret = loop.run_until_complete(coro)
        loop.close()

        if warns:
            raise RuntimeError

        return ret


def awaiter(func):
    def sync_func(*args, **kwargs):
        return _await(func(*args, **kwargs))

    return sync_func



# Copyright (c) 2021 Jeff Irion and contributors
#
# This file is part of the adb-shell package.  It incorporates work
# covered by the following license notice:
#
#
#   Copyright 2014 Google Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Implement helpers for the :class:`~adb_shell.adb_device.AdbDevice` and :class:`~adb_shell.adb_device_async.AdbDeviceAsync` classes.

.. rubric:: Contents

* :class:`_AdbPacketStore`

    * :meth:`_AdbPacketStore.__contains__`
    * :meth:`_AdbPacketStore.__len__`
    * :meth:`_AdbPacketStore.clear`
    * :meth:`_AdbPacketStore.clear_all`
    * :meth:`_AdbPacketStore.find`
    * :meth:`_AdbPacketStore.find_allow_zeros`
    * :meth:`_AdbPacketStore.get`
    * :meth:`_AdbPacketStore.put`

* :class:`_AdbTransactionInfo`

    * :meth:`_AdbTransactionInfo.args_match`

* :class:`_FileSyncTransactionInfo`

    * :meth:`_FileSyncTransactionInfo.can_add_to_send_buffer`

* :func:`get_banner`
* :func:`get_files_to_push`

"""


from collections import namedtuple
from io import BytesIO
import os
import socket
import struct

try:
    from asyncio import Queue
except ImportError:  # pragma: no cover
    try:
        from queue import Queue
    except ImportError:
        from Queue import Queue

from . import constants


DeviceFile = namedtuple('DeviceFile', ['filename', 'mode', 'size', 'mtime'])


def get_files_to_push(local_path, device_path):
    """Get a list of the file(s) to push.

    Parameters
    ----------
    local_path : str
        A path to a local file or directory
    device_path : str
        A path to a file or directory on the device

    Returns
    -------
    local_path_is_dir : bool
        Whether or not ``local_path`` is a directory
    local_paths : list[str]
        A list of the file(s) to push
    device_paths : list[str]
        A list of destination paths on the device that corresponds to ``local_paths``

    """
    local_path_is_dir = not isinstance(local_path, BytesIO) and os.path.isdir(local_path)
    local_paths = [local_path] if not local_path_is_dir else os.listdir(local_path)
    device_paths = [device_path] if not local_path_is_dir else [device_path + '/' + f for f in local_paths]

    return local_path_is_dir, local_paths, device_paths


def get_banner():
    """Get the ``banner`` that will be signed in :meth:`adb_shell.adb_device.AdbDevice.connect` / :meth:`adb_shell.adb_device_async.AdbDeviceAsync.connect`.

    Returns
    -------
    bytearray
        The hostname, or "unknown" if it could not be determined

    """
    try:
        return bytearray(socket.gethostname(), 'utf-8')
    except:  # noqa pylint: disable=bare-except
        return bytearray('unknown', 'utf-8')


class _AdbTransactionInfo(object):  # pylint: disable=too-few-public-methods
    """A class for storing info and settings used during a single ADB "transaction."

    Note that if ``timeout_s`` is not ``None``, then:

    ::

       self.transport_timeout_s <= self.read_timeout_s <= self.timeout_s

    If ``timeout_s`` is ``None``, the first inequality still applies.


    Parameters
    ----------
    local_id : int
        The ID for the sender (i.e., the device running this code)
    remote_id : int
        The ID for the recipient
    transport_timeout_s : float, None
        Timeout in seconds for sending and receiving data, or ``None``; see :meth:`BaseTransport.bulk_read() <adb_shell.transport.base_transport.BaseTransport.bulk_read>`,
        :meth:`BaseTransport.bulk_write() <adb_shell.transport.base_transport.BaseTransport.bulk_write>`,
        :meth:`BaseTransportAsync.bulk_read() <adb_shell.transport.base_transport_async.BaseTransportAsync.bulk_read>`, and
        :meth:`BaseTransportAsync.bulk_write() <adb_shell.transport.base_transport_async.BaseTransportAsync.bulk_write>`
    read_timeout_s : float
        The total time in seconds to wait for data and packets from the device
    timeout_s : float, None
        The total time in seconds to wait for the ADB command to finish

    Attributes
    ----------
    local_id : int
        The ID for the sender (i.e., the device running this code)
    read_timeout_s : float
        The total time in seconds to wait for data and packets from the device
    remote_id : int
        The ID for the recipient
    timeout_s : float, None
        The total time in seconds to wait for the ADB command to finish
    transport_timeout_s : float, None
        Timeout in seconds for sending and receiving data, or ``None``; see :meth:`BaseTransport.bulk_read() <adb_shell.transport.base_transport.BaseTransport.bulk_read>`,
        :meth:`BaseTransport.bulk_write() <adb_shell.transport.base_transport.BaseTransport.bulk_write>`,
        :meth:`BaseTransportAsync.bulk_read() <adb_shell.transport.base_transport_async.BaseTransportAsync.bulk_read>`, and
        :meth:`BaseTransportAsync.bulk_write() <adb_shell.transport.base_transport_async.BaseTransportAsync.bulk_write>`

    """
    def __init__(self, local_id, remote_id, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S, timeout_s=None):
        self.local_id = local_id
        self.remote_id = remote_id
        self.timeout_s = timeout_s
        self.read_timeout_s = read_timeout_s if self.timeout_s is None else min(read_timeout_s, self.timeout_s)
        self.transport_timeout_s = self.read_timeout_s if transport_timeout_s is None else min(transport_timeout_s, self.read_timeout_s)

    def args_match(self, arg0, arg1, allow_zeros=False):
        """Check if ``arg0`` and ``arg1`` match this object's ``remote_id`` and ``local_id`` attributes, respectively.

        Parameters
        ----------
        arg0 : int
            The ``arg0`` value from an ADB packet, which will be compared to this object's ``remote_id`` attribute
        arg1 : int
            The ``arg1`` value from an ADB packet, which will be compared to this object's ``local_id`` attribute
        allow_zeros : bool
            Whether to check if ``arg0`` and ``arg1`` match 0, in addition to this object's ``local_id`` and ``remote_id`` attributes

        Returns
        -------
        bool
            Whether ``arg0`` and ``arg1`` match this object's ``local_id`` and ``remote_id`` attributes

        """
        if not allow_zeros:
            return arg1 == self.local_id and (self.remote_id is None or arg0 == self.remote_id)

        # https://github.com/JeffLIrion/adb_shell/blob/17540be9b3b84637aca9b994ae3e0b35d02b1a03/adb_shell/adb_device.py#L923-L929
        return arg1 in (0, self.local_id) and (self.remote_id is None or arg0 in (0, self.remote_id))


class _FileSyncTransactionInfo(object):  # pylint: disable=too-few-public-methods
    """A class for storing info used during a single FileSync "transaction."

    Parameters
    ----------
    recv_message_format : bytes
        The FileSync message format
    maxdata: int
        Maximum amount of data in an ADB packet

    Attributes
    ----------
    _maxdata: int
        Maximum amount of data in an ADB packet
    recv_buffer : bytearray
        A buffer for storing received data
    recv_message_format : bytes
        The FileSync message format
    recv_message_size : int
        The FileSync message size
    send_buffer : bytearray
        A buffer for storing data to be sent
    send_idx : int
        The index in ``recv_buffer`` that will be the start of the next data packet sent

    """
    def __init__(self, recv_message_format, maxdata=constants.MAX_ADB_DATA):
        self.send_buffer = bytearray(maxdata)
        self.send_idx = 0

        self.recv_buffer = bytearray()
        self.recv_message_format = recv_message_format
        self.recv_message_size = struct.calcsize(recv_message_format)

        self._maxdata = maxdata

    def can_add_to_send_buffer(self, data_len):
        """Determine whether ``data_len`` bytes of data can be added to the send buffer without exceeding :const:`constants.MAX_ADB_DATA`.

        Parameters
        ----------
        data_len : int
            The length of the data to be potentially added to the send buffer (not including the length of its header)

        Returns
        -------
        bool
            Whether ``data_len`` bytes of data can be added to the send buffer without exceeding :const:`constants.MAX_ADB_DATA`

        """
        added_len = self.recv_message_size + data_len
        return self.send_idx + added_len < self._maxdata


class _AdbPacketStore(object):
    """A class for storing ADB packets.

    This class is used to support multiple streams.

    Attributes
    ----------
    _dict : dict[int: dict[int: Queue]]
        A dictionary of dictionaries of queues.  The first (outer) dictionary keys are the ``arg1`` return values from
        the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and
        :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods.  The second (inner)
        dictionary keys are the ``arg0`` return values from those methods.  And the values of this inner dictionary are
        queues of ``(cmd, data)`` tuples.

    """

    def __init__(self):
        self._dict = {}

    def __contains__(self, value):
        """Check if there are any entries in a queue for the specified value.

        Note that ``None`` is used as a wildcard.

        Parameters
        ----------
        value : tuple[int, int]
            An ``(arg0, arg1)`` pair; either or both values can be ``None``

        Returns
        -------
        bool
            Whether the ``(arg0, arg1)`` tuple has any corresponding queue entries

        """
        return bool(self.find(value[0], value[1]))

    def __len__(self):
        """Get the number of non-empty queues.

        Returns
        -------
        int
            The number of non-empty queues

        """
        return sum(not val0.empty() for val1 in self._dict.values() for val0 in val1.values())

    def clear(self, arg0, arg1):
        """Delete the entry for ``(arg0, arg1)``, if it exists.

        Parameters
        ----------
        arg0 : int
            The ``arg0`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods
        arg1 : int
            The ``arg1`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods

        """
        if arg1 in self._dict and arg0 in self._dict[arg1]:
            del self._dict[arg1][arg0]

            if not self._dict[arg1]:
                # `self._dict[arg1]` is an empty dictionary now, so delete it
                del self._dict[arg1]

    def clear_all(self):
        """Clear all the entries."""
        self._dict = {}

    def find(self, arg0, arg1):
        """Find the entry corresponding to ``arg0`` and ``arg1``.

        Parameters
        ----------
        arg0 : int, None
            The ``arg0`` value that we are looking for; ``None`` serves as a wildcard
        arg1 : int, None
            The ``arg1`` value that we are looking for; ``None`` serves as a wildcard

        Returns
        -------
        tuple[int, int], None
            The ``(arg0, arg1)`` pair that was found in the dictionary of dictionaries, or ``None`` if no match was found

        """
        if not self._dict:
            return None

        if arg1 is None:
            if arg0 is None:
                # `value = (None, None)` -> search for any non-empty queue
                return next(((key0, key1) for key1, val1 in self._dict.items() for key0, val0 in val1.items() if not val0.empty()), None)

            # Search for a non-empty queue with a key of `arg0 == value[0]`
            return next(((arg0, key1) for key1, val1 in self._dict.items() for key0, val0 in val1.items() if key0 == arg0 and not val0.empty()), None)

        if arg1 not in self._dict:
            return None

        if arg0 is None:
            # Look for a non-empty queue in the `self._dict[value[1]]` dictionary
            return next(((key0, arg1) for key0, val0 in self._dict[arg1].items() if not val0.empty()), None)

        if arg0 in self._dict[arg1] and not self._dict[arg1][arg0].empty():
            return (arg0, arg1)

        return None

    def find_allow_zeros(self, arg0, arg1):
        """Find the entry corresponding to (``arg0`` or 0) and (``arg1`` or 0).

        Parameters
        ----------
        arg0 : int, None
            The ``arg0`` value that we are looking for; ``None`` serves as a wildcard
        arg1 : int, None
            The ``arg1`` value that we are looking for; ``None`` serves as a wildcard

        Returns
        -------
        tuple[int, int], None
            The first matching ``(arg0, arg1)`` pair that was found in the dictionary of dictionaries, or ``None`` if no match was found

        """
        for arg0_, arg1_ in ((arg0, arg1), (arg0, 0), (0, arg1), (0, 0)):
            arg0_arg1 = self.find(arg0_, arg1_)
            if arg0_arg1:
                return arg0_arg1

        return None

    def get(self, arg0, arg1):
        """Get the next entry from the queue for ``arg0`` and ``arg1``.

        This function assumes you have already checked that ``(arg0, arg1) in self``.

        Parameters
        ----------
        arg0 : int, None
            The ``arg0`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods; ``None`` serves as a wildcard
        arg1 : int, None
            The ``arg1`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods; ``None`` serves as a wildcard

        Returns
        -------
        cmd : bytes
            The ADB packet's command
        arg0 : int
            The ``arg0`` value from the returned packet
        arg1 : int
            The ``arg1`` value from the returned packet
        data : bytes
            The ADB packet's data

        """
        if arg0 is None or arg1 is None:
            arg0, arg1 = self.find(arg0, arg1)

        # Get the data from the queue
        cmd, data = self._dict[arg1][arg0].get_nowait()

        # If this is a `CLSE` packet, then clear the entry in the store
        if cmd == constants.CLSE:
            self.clear(arg0, arg1)

        return cmd, arg0, arg1, data

    def put(self, arg0, arg1, cmd, data):
        """Add an entry to the queue for ``arg0`` and ``arg1``.

        Note that a new dictionary entry will not be created if ``cmd == constants.CLSE``.

        Parameters
        ----------
        arg0 : int
            The ``arg0`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods
        arg1 : int
            The ``arg1`` return value from the :meth:`adb_shell.adb_device._AdbIOManager._read_packet_from_device` and :meth:`adb_shell.adb_device_async._AdbIOManagerAsync._read_packet_from_device` methods
        cmd : bytes
            The ADB packet's command
        data : bytes
            The ADB packet's data

        """
        if arg1 in self._dict:
            if arg0 not in self._dict[arg1]:
                if cmd == constants.CLSE:
                    return

                # Create the `arg0` entry in the `arg1` dict
                self._dict[arg1][arg0] = Queue()
        else:
            if cmd == constants.CLSE:
                return

            # Create the `arg1` entry with a new dict
            self._dict[arg1] = {arg0: Queue()}

        # Put the data into the queue
        self._dict[arg1][arg0].put_nowait((cmd, data))



import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from adb_shell.auth.keygen import get_user_info


class TestKeygen(unittest.TestCase):
    def test_get_user_info(self):
        with patch('adb_shell.auth.keygen.os.getlogin', side_effect=OSError), patch('adb_shell.auth.keygen.socket.gethostname', return_value=''):
            user_host = get_user_info()
            self.assertEqual(user_host, ' unknown@unknown')

        with patch('adb_shell.auth.keygen.os.getlogin', return_value=''), patch('adb_shell.auth.keygen.socket.gethostname', return_value=''):
            user_host = get_user_info()
            self.assertEqual(user_host, ' unknown@unknown')



from contextlib import contextmanager
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class FileReadWrite(object):
    """Mock an opened file that can be read and written to."""
    def __init__(self):
        self._content = b''
        self._mode = 'r'

    def read(self):
        if self._mode == 'r':
            if not isinstance(self._content, str):
                return self._content.decode()
            return self._content

        if isinstance(self._content, str):
            return self._content.encode('utf-8')
        return self._content
        

    def write(self, content):
        self._content = content


PRIVATE_KEY = FileReadWrite()
PUBLIC_KEY = FileReadWrite()


@contextmanager
def open_priv_pub(infile, mode='r'):
    try:
        if infile.endswith('.pub'):
            PUBLIC_KEY._mode = mode
            yield PUBLIC_KEY
        else:
            PRIVATE_KEY._mode = mode
            yield PRIVATE_KEY
    finally:
        pass



try:
    from contextlib import asynccontextmanager
except ImportError:
    asynccontextmanager = lambda func: func

from unittest.mock import patch

from adb_shell import constants
from adb_shell.adb_message import AdbMessage, unpack
from adb_shell.transport.tcp_transport_async import TcpTransportAsync

try:
    from unittest.mock import AsyncMock
except ImportError:
    from unittest.mock import MagicMock

    class AsyncMock(MagicMock):
        async def __call__(self, *args, **kwargs):
            return super(AsyncMock, self).__call__(*args, **kwargs)


def async_mock_open(read_data=""):
    class AsyncMockFile:
        def __init__(self, read_data):
            self.read_data = read_data
            _async_mock_open.written = read_data[:0]

        async def read(self, size=-1):
            if size == -1:
                ret = self.read_data
                self.read_data = self.read_data[:0]
                return ret

            n = min(size, len(self.read_data))
            ret = self.read_data[:n]
            self.read_data = self.read_data[n:]
            return ret

        async def write(self, b):
            if _async_mock_open.written:
                _async_mock_open.written += b
            else:
                _async_mock_open.written = b

        def fileno(self):
            return 123

    @asynccontextmanager
    async def _async_mock_open(*args, **kwargs):
        try:
            yield AsyncMockFile(read_data)
        finally:
            pass

    return _async_mock_open


class FakeStreamWriter:
    def close(self):
        pass

    async def wait_closed(self):
        pass

    def write(self, data):
        pass

    async def drain(self):
        pass


class FakeStreamReader:
    async def read(self, numbytes):
        return b'TEST'


class FakeTcpTransportAsync(TcpTransportAsync):
    def __init__(self, *args, **kwargs):
        TcpTransportAsync.__init__(self, *args, **kwargs)
        self.bulk_read_data = b''
        self.bulk_write_data = b''

    async def close(self):
        self._reader = None
        self._writer = None

    async def connect(self, transport_timeout_s=None):
        self._reader = True
        self._writer = True

    async def bulk_read(self, numbytes, transport_timeout_s=None):
        num = min(numbytes, constants.MAX_ADB_DATA)
        ret = self.bulk_read_data[:num]
        self.bulk_read_data = self.bulk_read_data[num:]
        return ret

    async def bulk_write(self, data, transport_timeout_s=None):
        self.bulk_write_data += data
        return len(data)


# `TcpTransport` patches
PATCH_TCP_TRANSPORT_ASYNC = patch('adb_shell.adb_device_async.TcpTransportAsync', FakeTcpTransportAsync)


def async_patch(*args, **kwargs):
    return patch(*args, new_callable=AsyncMock, **kwargs)



"""setup.py file for the adb_shell package."""

from setuptools import setup

with open("README.rst") as f:
    readme = f.read()

setup(
    name="adb_shell",
    version="0.4.4",
    description="A Python implementation of ADB with shell and FileSync functionality.",
    long_description=readme,
    keywords=["adb", "android"],
    url="https://github.com/JeffLIrion/adb_shell",
    author="Jeff Irion",
    author_email="jefflirion@users.noreply.github.com",
    packages=["adb_shell", "adb_shell.auth", "adb_shell.transport"],
    install_requires=["cryptography", "pyasn1", "rsa"],
    tests_require=["pycryptodome", "libusb1>=1.0.16"],
    extras_require={"usb": ["libusb1>=1.0.16"], "async": ["aiofiles>=0.4.0", "async_timeout>=3.0.0"]},
    classifiers=[
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
    ],
    test_suite="tests",
)



import asyncio
import unittest
from unittest.mock import patch

from adb_shell.exceptions import TcpTimeoutException
from adb_shell.transport.tcp_transport_async import TcpTransportAsync

from .async_patchers import FakeStreamReader, FakeStreamWriter, async_patch
from .async_wrapper import awaiter
from . import patchers


@patchers.ASYNC_SKIPPER
class TestTcpTransportAsync(unittest.TestCase):
    def setUp(self):
        """Create a ``TcpTransportAsync`` and connect to a TCP service.

        """
        self.transport = TcpTransportAsync('host', 5555)

    @awaiter
    async def test_close(self):
        await self.transport.close()

    @awaiter
    async def test_close2(self):
        await self.transport.close()

    @awaiter
    async def test_connect(self):
        with async_patch('asyncio.open_connection', return_value=(True, True)):
            await self.transport.connect(transport_timeout_s=1)

    @awaiter
    async def test_connect_close(self):
        with async_patch('asyncio.open_connection', return_value=(FakeStreamReader(), FakeStreamWriter())):
            await self.transport.connect(transport_timeout_s=1)
            self.assertIsNotNone(self.transport._writer)

        await self.transport.close()
        self.assertIsNone(self.transport._reader)
        self.assertIsNone(self.transport._writer)

    @awaiter
    async def test_connect_close_catch_oserror(self):
        with async_patch('asyncio.open_connection', return_value=(FakeStreamReader(), FakeStreamWriter())):
            await self.transport.connect(transport_timeout_s=1)
            self.assertIsNotNone(self.transport._writer)

        with patch('{}.FakeStreamWriter.close'.format(__name__), side_effect=OSError):
            await self.transport.close()
            self.assertIsNone(self.transport._reader)
            self.assertIsNone(self.transport._writer)

    @awaiter
    async def test_connect_with_timeout(self):
        with self.assertRaises(TcpTimeoutException):
            with async_patch('asyncio.open_connection', side_effect=asyncio.TimeoutError):
                await self.transport.connect(transport_timeout_s=1)

    @awaiter
    async def test_bulk_read(self):
        with async_patch('asyncio.open_connection', return_value=(FakeStreamReader(), FakeStreamWriter())):
            await self.transport.connect(transport_timeout_s=1)

        self.assertEqual(await self.transport.bulk_read(4, transport_timeout_s=1), b'TEST')

        with self.assertRaises(TcpTimeoutException):
            with patch('{}.FakeStreamReader.read'.format(__name__), side_effect=asyncio.TimeoutError):
                await self.transport.bulk_read(4, transport_timeout_s=1)

    @awaiter
    async def test_bulk_write(self):
        with async_patch('asyncio.open_connection', return_value=(FakeStreamReader(), FakeStreamWriter())):
            await self.transport.connect(transport_timeout_s=1)

        self.assertEqual(await self.transport.bulk_write(b'TEST', transport_timeout_s=1), 4)

        with self.assertRaises(TcpTimeoutException):
            with patch('{}.FakeStreamWriter.write'.format(__name__), side_effect=asyncio.TimeoutError):
                await self.transport.bulk_write(b'TEST', transport_timeout_s=1)
